package com.blazinggames.retro.classic.search;

// Spelchan Imports
import com.blazinggames.urcl.data.Config;
import com.blazinggames.urcl.ui.*;
import com.blazinggames.urcl.text.*;

// Retro Imports
import com.blazinggames.retro.shared.search.*;

/**
 * The game display for the Tank classic version. Tank is a game where
 * two to nine players (human or computer controlled) move around a
 * two dimensional grid while taking pot shots at each other. The last
 * player standing wins.
 * ------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Billy D. Spelchan
 * @version 1.00.011016
*/
public class Tanks 
extends TextDisplay
{
// *********************
// ***** CONSTANTS *****
// *********************

    private static final String VERSION = "1.10.040101";

    private static final String INSTRUCTIONS = "09INSTRUCTIONS";
    private static final String MAX_PLAYERS = "09MAXPLAYERS";
    private static final String WIDTH = "09WIDTH";
    private static final String HEIGHT = "09HEIGHT";
    private static final String PLAYFIELDWRAP = "09WRAP";
    private static final String MAX_MOVE = "09MAXMOVE";
    private static final String MAX_TURN = "09MAXTURN";
    private static final String MAX_POWER = "09MAXPOWER";
    private static final String EXPLODE = "09EXPLODERADIUS";


// *********************
// ***** VARIABLES *****
// *********************

    private Instructions _instructions;

    // Game configuration
    private int _maxPlayers;
    private int _playfieldWidth;
    private int _playfieldHeight;
    private boolean _playfieldWraps;
    private int _maxMove;
    private int _maxTurn;
    private int _maxPower;
    private int _explodeRadius;
    
    // Game settings
    private int _numPlayers;
    private String _playerNames[];
    private boolean _playerIsHuman[];
    private boolean _playerBeginner[];
    private Tank _tanks[];

    private double _lastHitX, _lastHitY;
    // AI variables
    private int _aiLastTarget[];
//    private int _aiLastPower[];
//    private boolean _aiWasLastShotLong[];
    private int _aiLowPower[];
    private int _aiHighPower[];
    
// **************************************    
// ***** METHODS - From TextDisplay *****
// **************************************

    public boolean initialize(DisplayHandler dh)
    {
        super.initialize(dh);
        
    	_maxPlayers = _config.getIntProperty(MAX_PLAYERS, 2);
    	_playfieldWidth = _config.getIntProperty(WIDTH, 1000);
    	_playfieldHeight = _config.getIntProperty(HEIGHT, 1000);
    	_playfieldWraps = (_config.getIntProperty(PLAYFIELDWRAP, 0) == 1);
    	_maxMove = _config.getIntProperty(MAX_MOVE, 2);
    	_maxTurn = _config.getIntProperty(MAX_TURN, 45);
        _maxPower = _config.getIntProperty(MAX_POWER, 1000);
        _explodeRadius = _config.getIntProperty(EXPLODE, 1);
        
        return true;
    }


    public void run()
    {
        String s;

        printHeader();
        instructions();

        do
        {
            gameSettings();
            gameStart();
            gameLoop();
            
            if (_canQuit)
            {
                print("Do you want another game?");
                s = input();
                println(s);
                if ((s.startsWith("N")) || (s.startsWith("n")))
                    _displayHandler.displayEvent(this, "SHOW 1");
            }
        }
        while (_active);
    }

    
// *************************************
// ***** METHODS - Game management *****
// *************************************

    private void printHeader()
    {
        println("*** TANKS ***");
        println("Clasic version release " + VERSION);
        println();
        println("Copyright (c) 2002-2004");
        println("Blazing Games Inc. All Rights Reserved.");
        println();
    }

    private void instructions()
    {
        print("Do you need instructions?");
        String s = input();
        println(s);
        if ( (s.startsWith("y")) || (s.startsWith("Y")) )
        {
            if (_instructions == null)
            {
                println("Grabbing instructions...");
                _res.addTextGroup(INSTRUCTIONS, _config.getStringProperty(INSTRUCTIONS, "tank.txt"));
                _instructions = new Instructions(_res.getTextGroup(INSTRUCTIONS));
                println("Done.");
                println();
            }
            _instructions.showInstructions(this);
        }
    }

    public void gameSettings()
    {
        String s;
        
        if (_numPlayers > 1)
        {
            println("Keep same settings? ");
            s = input();
            if ( (s.startsWith("y")) || (s.startsWith("Y")) )
                return;
        }
        
        do
        {
            print("How many players (2 to "+_maxPlayers+")? ");
            s = input();
            _numPlayers = _config.parseInt(s, 0);
            println("" + _numPlayers);
        }
        while ( (_numPlayers < 2) || (_numPlayers > _maxPlayers) );
        
        int numAIs = 0;
        _playerNames = new String[_numPlayers];
        _playerIsHuman = new boolean[_numPlayers];
        _playerBeginner = new boolean[_numPlayers];
        _aiLastTarget = new int[_numPlayers];
//        _aiLastPower = new int[_numPlayers];
//        _aiWasLastShotLong = new boolean[_numPlayers];
        _aiLowPower = new int[_numPlayers];
        _aiHighPower = new int[_numPlayers];
        
        for (int cntr = 0; cntr < _numPlayers; ++cntr)
        {
            print("is player " + (cntr + 1) + " human? ");
            s = input();
            println(s);
            _playerIsHuman[cntr] = ( (s.startsWith("y")) || (s.startsWith("Y")) );
            if (_playerIsHuman[cntr])
            {
                print("Name: ");
                _playerNames[cntr] = input();
                println(_playerNames[cntr]);
                _playerBeginner[cntr] = yesNo("is player a beginner? ");
            }
            else
            {
                ++numAIs;
                _playerNames[cntr] = "Computer " + numAIs;
                println("Being named " + _playerNames[cntr]);
                _aiLastTarget[cntr] = -1;
            }
        }
    }
    
    public void gameStart()
    {
        _tanks = new Tank[_numPlayers];
        for (int cntr = 0; cntr < _numPlayers; ++cntr)
        {
            _tanks[cntr] = new Tank(rand(_playfieldWidth), rand(_playfieldHeight), 90);
        }
    }
    
    public void gameLoop()
    {
        int round = 0;
        int cntr;
        double px, py;
        
        do
        {
           ++ round;
            println();
            println("*** ROUND " + round + " ***");
            println();
            for (cntr = 0; cntr < _numPlayers; ++cntr)
            {
                if (_tanks[cntr].isAlive())
                {
                    px = (double)(Math.round(_tanks[cntr].getX() * 100)) / 100;
                    py = (double)(Math.round(_tanks[cntr].getY() * 100)) / 100;
                    println(_playerNames[cntr] + " is at " + px + ", " + py);
                }
            }

            for (cntr = 0; cntr < _numPlayers; ++cntr)
            {
                if (_tanks[cntr].isAlive())
                {
                    if (_playerIsHuman[cntr])
                        handleHumanMove(cntr);
                    else
                        handleComputerMove(cntr);
                }
            }
        }
        while (isGameOver() != true);
    }
    
    public boolean isGameOver()
    {
        int humanCount = 0;
        int tankCount = 0;
        int lastLiving = 0;
        
        for (int cntr = 0; cntr < _numPlayers; ++cntr)
        {
            if (_tanks[cntr].isAlive())
            {
                ++tankCount;
                if (_playerIsHuman[cntr])
                    ++humanCount;
                lastLiving = cntr;
            }
        }

        if (tankCount < 2)
        {
            println();
            println("*** GAME OVER ***");
            println();
            if (tankCount == 0)
                println("Everyone is gone!");
            else
                println(_playerNames[lastLiving] + " has won the game!");
            println();
            return true;
        }

        if (humanCount < 1)
        {
            println();
            println("*** GAME OVER ***");
            println();            
            println("All humans have been defeated!");
            println();
            return true;
        }

        return false;
    }
    
    public void handleHumanMove(int id)
    {
        int turn, advance, fdir, fmin, fsec, fang, fpow;
        String s;
        double px = (double)(Math.round(_tanks[id].getX() * 100)) / 100;
        double py = (double)(Math.round(_tanks[id].getY() * 100)) / 100;

        println();
        println("It is " + _playerNames[id] + " turn!");
/*        
        println("You are at " + px + ", " + py + " facing " + _tanks[id].getDirection());        
        turn = requestRange("Turn (-" + _maxTurn + " to " + _maxTurn + "): ", -_maxTurn, _maxTurn);
        _tanks[id].turn(turn);
        advance = requestRange("Advance (-1 to " + _maxMove + "): ", -1, _maxMove);
        _tanks[id].move(advance);
        px = (double)(Math.round(_tanks[id].getX() * 100)) / 100;
        py = (double)(Math.round(_tanks[id].getY() * 100)) / 100;
        println("Now at " + px + ", " + py + " facing " + _tanks[id].getDirection());
*/
        if (_playerBeginner[id])
            printAngleChart(id);
        
        fdir = requestRange("Direction to fire (degrees): ", 0, 359);
        fmin = requestRange("Direction to fire (minutes): ", 0, 59);
        fsec = requestRange("Direction to fire (seconds): ", 0, 59);
        _tanks[id].setFireDirection(fdir, fmin, fsec);
        fang = requestRange("Turret angle (0 to 90): ", 0, 90);
        _tanks[id].setFireAngle(fang);
        fpow = requestRange("Power (1 to " + _maxPower + "): ", 1, _maxPower);
        _tanks[id].setFirePower(fpow);
        
        handleProjectile(_tanks[id].fire(), yesNo("Show bullet path (y/n)? "));
    }
    
    private void handleProjectile(Projectile p, boolean showPath)
    {
        boolean alive = true;
        double hitX, hitY, hitHeight;
        double px, py;
        double temp, distx, disty, dist;
        double erad = _explodeRadius * _explodeRadius;
        
        // move projectile
        do
        {
            alive = p.tick(1000);
//            hitX = (double)(Math.round(p.getX() * 100)) / 100;
//            hitY = (double)(Math.round(p.getY() * 100)) / 100;
//            hitHeight = (double)(Math.round(p.getHeight() * 100)) / 100;
            hitX = p.getX();
            hitY = p.getY();
            hitHeight = p.getHeight();
            if (showPath)
                println("@" + floatString( hitX ) + "," + floatString(hitY) + " height " + floatString(hitHeight));
        }
        while (alive);
        println("Projectile hit at " + floatString(hitX) + ", " + floatString(hitY));
        _lastHitX = hitX;
        _lastHitY = hitY;
        
        // calculate who is in blast radius (by comparing distance squareds
//System.out.println("projectile radius squared is " + erad);
        for (int cntr = 0; cntr < _numPlayers; ++cntr)
        {
            if (_tanks[cntr].isAlive())
            {
                px = (double)(Math.round(_tanks[cntr].getX() * 100)) / 100;
                py = (double)(Math.round(_tanks[cntr].getY() * 100)) / 100;
                temp = hitX - px;
                distx = temp * temp;
                temp = hitY - py;
                disty = temp * temp;
                dist = distx + disty;
                
//System.out.println(_playerNames[cntr] + " is sqrt(" + dist + ") from blast");
                if (dist <= erad)
                {
                    println("* Tank Destroyed! *");
                    _tanks[cntr].setAlive(false);
                    println(_playerNames[cntr] + " has been destroyed!");
                }
            }
        }
        
    }
    
    private int requestRange(String querry, int low, int high)
    {
        int result;
        String s;
        
        do
        {
            print(querry);
            s = input();
            println(s);
            result = _config.parseInt(s, high + 1);
        }
        while ( (result < low) || (result > high) );
        
        return result;
    }

    private boolean yesNo(String prompt)
    {
        String s;
        boolean valid = false;
        boolean result = false;

        do
        {
            print(prompt);
            s = input();
            println(s);
            valid = true;
            if ( (s.startsWith("Y")) || (s.startsWith("y")) )
                result = true;
            else if ( (s.startsWith("N")) || (s.startsWith("n")) )
                result = false;
            else
                valid = false;
        }
        while ( !valid );

        return result;
    }


// ********************************
// ***** METHODS: AI Handling *****
// ********************************
    
    public void handleComputerMove(int id)
    {
        println();
        println(_playerNames[id] + " is shooting");
//        aiMovement(id);
        aiTargeting(id);
    }

    private void printAngleChart(int id)
    {
        double tempAng;
        int n, m, s;
        double dtemp;
        
        println("Beginnner Angle chart for " + _playerNames[id]);
        for (int cntr = 0; cntr < _numPlayers; ++cntr)
        {
            if ( (cntr != id) && (_tanks[cntr].isAlive()) )
            {
                tempAng = findAngle(id, cntr);
                n = (int)tempAng;
                dtemp = tempAng - (double)n;
                m = (int)(dtemp * 60);
                s = (int)((dtemp * 3600) % 60);
                println(" " + _playerNames[cntr] + " is at " + n + "d " + m + "m " + s + "s" );
            }
        }
    }
    
    private double findAngle(int id, int target)
    {
        // s = source, t = target, d = delta
        double sx = _tanks[id].getX();
        double tx = _tanks[target].getX();
        double dx = Math.abs(sx - tx);
        double sy = _tanks[id].getY();
        double ty = _tanks[target].getY();
        double dy = Math.abs(sy - ty);
        double tanAng = 0;

        if (dx > 0)
            tanAng = dy / dx;
        else
            tanAng = dy / 0.000000001;
        double angRad = Math.atan(tanAng);
        double angDeg = angRad * 180 / Math.PI;

        if (sx < tx)
        {
            if (sy < ty) // quad I
                return angDeg;
            else // quad IV
                return 360 - angDeg;
        }
        else
        {
            if (sy < ty) // quad II
                return 180 - angDeg;
            else // quad III
                return 180 + angDeg;
        }
    }

/*
    public void aiMovement(int id)
    {
        int turnAngle = rand(-60, 60);
        int advAmount = rand(-1, 4);

        _tanks[id].turn(turnAngle);
        _tanks[id].move(advAmount);
        double px = (double)(Math.round(_tanks[id].getX() * 100)) / 100;
        double py = (double)(Math.round(_tanks[id].getY() * 100)) / 100;
        println(_playerNames[id] + " Now at " + px + ", " + py + " facing " + _tanks[id].getDirection());
    }
*/
    
    public void aiTargeting(int id)
    {
        // first see if need to aquire new target
        if (_aiLastTarget[id] == -1)
            aiAquireTarget(id);
        else if (_tanks[_aiLastTarget[id]].isAlive() == false)
            aiAquireTarget(id);
        
        // Next calculate targetting angle (computer cheats by getting new angle every turn)
        double angle = findAngle(id, _aiLastTarget[id]);
        int power, powerAdj;
        
        // Third, calculate new power to use
/*
        if (_aiWasLastShotLong[id])
        {
            powerAdj = rand(50);
            power = _aiLastPower[id] - powerAdj;
        }
        else
        {
            powerAdj = rand(100);
            power = _aiLastPower[id] + powerAdj;            
        }
*/
        power = rand(_aiLowPower[id], _aiHighPower[id]);
        
        // Finally take the shot and update ai info
        double tempX = _tanks[id].getX() - _tanks[_aiLastTarget[id]].getX();
        double tempY = _tanks[id].getY() - _tanks[_aiLastTarget[id]].getY();
        double targDist = Math.sqrt(tempX * tempX + tempY*tempY);
        _tanks[id].setFireDirection(angle);
        _tanks[id].setFireAngle(45);
        _tanks[id].setFirePower(power);
//System.out.println(id + " is firing at power " + power);
        handleProjectile(_tanks[id].fire(), false);
        tempX = _lastHitX - _tanks[id].getX();
        tempY = _lastHitY - _tanks[id].getY();
        double hitDist = Math.sqrt(tempX * tempX + tempY * tempY);
//        _aiLastPower[id] = power;
        if (targDist < hitDist)
            _aiHighPower[id] = power;
        else
            _aiLowPower[id] = power;
    }

    public void aiAquireTarget(int id)
    {
        // build potential valid list
        int potentialList[] = new int[_numPlayers];
        int cntr, valid = 0, pick;

        for (cntr = 0; cntr < _numPlayers; ++cntr)
        {
            if ( (id != cntr) && (_tanks[cntr].isAlive()))
            {
                potentialList[valid] = cntr;
                ++valid;
            }
        }

        // pick from list (or 0 if empty list)
        if (valid > 0)
        {
            pick = potentialList[rand(valid) - 1];
        }
        else
            pick = 0;
        _aiLastTarget[id] = pick;
//        _aiWasLastShotLong[id] = false;
//        _aiLastPower[id] = 50;
        _aiHighPower[id] = 1000;
        _aiLowPower[id] = 0;
//System.out.println("AI for " + id + " Choosing " + pick + " as new target");
    }

    private String floatString(double d)
    {
        double temp = d + 0.0005;
        int left = (int)temp;
        temp -= (float)left;
        temp *= 1000;
        int right = (int)temp;
        StringBuffer rs = new StringBuffer();
        rs.append(left);
        rs.append(".");
        int targetlen = rs.length() + 3;
        rs.append(right);
        while(rs.length() < targetlen)
            rs.append("0");
        return rs.toString();
    }
    
}
